"""
Base functions for managing spec files
"""

import jinja2
import yaml

DEFAULT_CONTENT_TYPE = "application/json"

JINJA_ENV = jinja2.Environment(
    loader=jinja2.PackageLoader("firestone", package_path="schema"),
    autoescape=jinja2.select_autoescape(),
    extensions=["jinja2.ext.loopcontrols"],
)


class SchemaMissingAttribute(Exception):
    """Schema is missing an attribute."""


def yaml_pretty(data, indent=2):
    """A simple YAML pretty print for Jinja."""
    dump = yaml.dump(data, indent=2)
    res = ""
    for line in dump.split("\n"):
        res += " " * indent + line + "\n"
    return res.rstrip()


JINJA_ENV.filters["yaml_pretty"] = yaml_pretty


def get_opid(path: str, method: str):
    """Get a unique operationId given the path and method."""
    opid = path[1:].replace("/", "_")
    opid = opid.replace("{", "")
    opid = opid.replace("}", "")
    return f"{opid}_{method}"



"""
The main entry point for firestone.
"""

import logging

import click

from firestone_lib import cli as firestone_cli
from firestone_lib import resource as firestone_rsrc

from firestone import spec as firestone_spec

_LOGGER = logging.getLogger(__name__)


@click.group()
@click.option("--debug", help="Turn on debugging", is_flag=True)
def main(debug):
    """Main entry point"""
    firestone_cli.init_logging("firestone_lib.resources.logging", "cli.conf")

    if debug:
        logging.getLogger().setLevel(logging.DEBUG)
        logging.getLogger("firestone").setLevel(logging.DEBUG)


@main.group()
@click.option(
    "--description",
    "-d",
    help="The description of this project",
    required=True,
)
@click.option(
    "--resources",
    "-r",
    help="One or more resource files in JSON Schema format, can be JSON or YAML",
    required=True,
    type=firestone_cli.PathList,
)
@click.option(
    "--summary",
    "-s",
    help="The summary of this project, defaults to description",
)
@click.option(
    "--title",
    "-t",
    help="The title of this project",
    required=True,
)
@click.option(
    "--version",
    "-v",
    help="The overal version of this spec",
    required=True,
)
@click.pass_context
def generate(ctx, description, resources, summary, title, version):
    """Upper command for gathering common resource information for the generators."""
    ctx.obj = {
        "data": [],
        "desc": description,
        "summary": summary,
        "title": title,
        "version": version,
    }
    for rsrc in resources:
        _LOGGER.debug(f"rsrc: {rsrc}")
        rsrc_data = firestone_rsrc.get_resource_schema(rsrc)
        _LOGGER.debug(f"rsrc_data: {rsrc_data}")
        _LOGGER.info(f"Validating resource {rsrc_data['name']} against firestone JSON schema.")
        firestone_rsrc.validate(rsrc_data)

        ctx.obj["data"].append(rsrc_data)


# TODO add support for providing an existing openapi spec file and merge data in
@generate.command()
@click.option(
    "--output",
    "-O",
    help="Output the specificaton to file name provided, or `-` for stdout",
    type=click.File("w"),
    default="-",
)
@click.option(
    "--ui-server",
    help="Launch web server to vierw Swagger UI",
    is_flag=True,
)
@click.option(
    "--prefix",
    help="A prefix to all URLs, this will add a 'servers' section to the openapi spec doc",
)
@click.option(
    "--security",
    help=(
        "Add security scheme to schema; e.g.: "
        '{"name": "bearer_auth", "scheme": "bearer", "type": "http", "bearerFormat": "JWT"}'
    ),
    type=firestone_cli.StrDict,
)
@click.option(
    "--version",
    help="Set the OpenAPI spec version",
    show_default=True,
    default=firestone_spec.openapi.DEFAULT_VERSION,
)
@click.pass_obj
def openapi(rsrc_data, output, ui_server, prefix, security, version):
    """Generate an OpenAPI specification for the given resource data."""

    openapi_spec = firestone_spec.openapi.generate(
        rsrc_data["data"],
        rsrc_data["title"],
        rsrc_data["desc"],
        rsrc_data["summary"],
        rsrc_data["version"],
        prefix=prefix,
        security=security,
        openapi_version=version,
    )
    print(openapi_spec, file=output)

    if ui_server:
        # pylint: disable=import-outside-toplevel,import-error,no-member
        import quart
        import swagger_ui

        app = quart.Quart(__name__)
        swagger_ui.quart_api_doc(
            app, config_spec=openapi_spec, url_prefix="/apidocs", title="OpenAPI doc"
        )

        app.run()


@generate.command()
@click.option(
    "--output",
    "-O",
    help="Output the specificaton to file name provided, or `-` for stdout",
    type=click.File("w"),
    default="-",
)
@click.pass_obj
def asyncapi(rsrc_data, output):
    """Generate an AsyncAPI specification for the given resource data."""

    asyncapi_spec = firestone_spec.asyncapi.generate(
        rsrc_data["data"],
        rsrc_data["title"],
        rsrc_data["desc"],
        rsrc_data["summary"],
        rsrc_data["version"],
    )
    print(asyncapi_spec, file=output)


@generate.command()
@click.option(
    "--output",
    "-O",
    help="Location of the main CLI generated file name, or `-` for stdout",
    type=click.File("w"),
    default="-",
    show_default=True,
)
@click.option(
    "--pkg",
    help="The package where the OpenAPI client code is.",
    required=True,
)
@click.option(
    "--client-pkg",
    help="The package where the OpenAPI client code is.",
    required=True,
)
@click.pass_obj
def cli(rsrc_data, pkg, client_pkg, output):
    """Generate python, Click-based CLI script.

    This generated script can be used as standalone or added to console scripts."""
    cli_spec = firestone_spec.cli.generate(
        pkg,
        client_pkg,
        rsrc_data["data"],
        rsrc_data["title"],
        rsrc_data["desc"],
        rsrc_data["summary"],
        rsrc_data["version"],
    )
    print(cli_spec, file=output)


if __name__ == "main":
    # pylint: disable=no-value-for-parameter
    main()



"""
Generate AsyncAPI 2.5 Spec
"""

# TODO: fix dupe code
# pylint: disable=duplicate-code

import enum
import copy
import logging

from firestone.spec import _base as spec_base

_LOGGER = logging.getLogger(__name__)


class OperationType(enum.Enum):
    """The operation type for a channel."""

    SUBSCRIBE = "subscribe"
    PUBLISH = "publish"


class SupportedBindingType(enum.Enum):
    """The supported binding type for an operation."""

    HTTP = "http"
    WS = "ws"


def get_payload(
    schema: dict,
    comp_name: str = None,
    attr_name: bool = None,
    is_list: bool = None,
):
    """Set schema for a given operation type."""
    # Default to using the schema directly in the file
    payload = schema["items"] if "items" in schema else schema

    # if a component name is provided, use that to reference it
    ref = None
    if comp_name:
        ref = f"#/components/schemas/{comp_name}"
    if attr_name:
        ref = f"#/components/schemas/{comp_name}/properties/{attr_name}"

    if is_list:
        items = {"$ref": ref} if ref else payload
        return {
            "type": "array",
            "items": items,
        }
    if ref:
        return {"$ref": ref}

    return payload


def get_message(
    rsrc_name: str,
    schema: dict,
    content_type: str,
    comp_name: str = None,
    attr_name: bool = None,
    is_list: bool = None,
):
    """Get a message for the operation.

    :param str rsrc_name: the resource name
    :param dict schema: the schema for this resource name
    :param str content_type: the content type for the messages
    """
    message = {"name": rsrc_name, "contentType": content_type}
    payload = get_payload(
        schema,
        comp_name=comp_name,
        attr_name=attr_name,
        is_list=is_list,
    )
    _LOGGER.debug(f"payload: {payload}")
    message["payload"] = payload

    return message


def get_binding(
    meta: dict,
    schema: dict,
    method: str,
):
    """Get a binding for the operation.

    :param dict meta: the meta data in the resource
    :param dict schema: the schema for this resource name
    :param str method: the method being used in this binging
    """
    query_params = schema.get("query_params", [])
    default_query_params = meta.get("default_query_params")
    query_params.extend(default_query_params)

    query = {}
    if "query_params" in schema:
        query = {"type": "object", "required": [], "properties": {}}
        for param in query_params:
            _LOGGER.debug(f"param: {param}")
            methods = param.get("methods", [])
            _LOGGER.debug(f"methods: {methods}")
            if methods:
                del param["methods"]
            if methods and method not in methods:
                continue

            param_name = param["name"]
            required = param.get("required", False)
            if required and param_name not in query["required"]:
                query["required"].append(param_name)

            query["properties"][param_name] = {
                "description": param["description"],
            }
            query["properties"][param_name].update(param.get("schema", {"type": "string"}))

    return {
        "method": method,
        "query": query,
    }


def get_channel(
    meta: dict,
    baseurl: str,
    schema: dict,
    descs: dict,
    keys: list = None,
    rsrc_name: str = None,
    attr_name: str = None,
    is_list: bool = None,
):
    """Get a channel for the given info.

    :param dict meta: the meta data in the resource
    :param str baseurl: the baseurl to use for channels
    :param dict schema: the schema for this resource name
    :param dict descs: the descriptions
    :param list keys: the keys dict for the instance of this resource
    :param bool is_list: this schema is for a list
    """

    # Sane default
    if is_list is None:
        is_list = False

    # TODO: send what operation types, as a list
    if not rsrc_name:
        rsrc_name = meta["name"]

    channel = {
        "description": f"Channel for {baseurl}",
        "parameters": {},
    }
    _LOGGER.debug(f"keys: {keys}")
    if keys:
        for key in keys:
            key_name = key["name"]
            _LOGGER.debug(f"key_name: {key_name}: {baseurl}")
            if key_name not in baseurl:
                continue
            channel["parameters"][key_name] = {
                "description": key["description"],
                "schema": key.get("schema") or {"type": "string"},
            }
    _LOGGER.debug(f"channel: {channel}")

    # 1. Add subscribers, i.e. get
    method = "get"
    message = get_message(
        rsrc_name,
        schema,
        spec_base.DEFAULT_CONTENT_TYPE,
        comp_name=rsrc_name,
        attr_name=attr_name,
        is_list=is_list,
    )

    binding = get_binding(meta, schema, method)
    _LOGGER.debug(f"binding: {binding}")
    channel[OperationType.SUBSCRIBE.value] = {
        "operationId": spec_base.get_opid(baseurl, OperationType.SUBSCRIBE.value),
        "description": descs.get(method, f"Subscribe from {baseurl}"),
        "message": message,
        "bindings": {
            SupportedBindingType.WS.value: binding,
        },
        "tags": [{"name": rsrc_name}],
    }

    # 2. Add publishers, i.e. post
    method = "post"
    message = get_message(
        rsrc_name,
        schema,
        spec_base.DEFAULT_CONTENT_TYPE,
        comp_name=rsrc_name,
        attr_name=attr_name,
    )
    binding = get_binding(meta, schema, method)
    _LOGGER.debug(f"binding: {binding}")

    channel[OperationType.PUBLISH.value] = {
        "operationId": spec_base.get_opid(baseurl, OperationType.PUBLISH.value),
        "description": descs.get(method, f"Publish to {baseurl}"),
        "message": message,
        "bindings": {
            SupportedBindingType.WS.value: binding,
        },
        "tags": [{"name": rsrc_name}],
    }

    return channel


def add_resource_level(
    meta: dict,
    baseurl: str,
    schema: dict,
    channels: dict,
    keys: list = None,
    rsrc_name: str = None,
):
    """Add resource level channels.

    :param dict meta: the meta data in the resource
    :param str baseurl: the baseurl to use for channels
    :param dict schema: the schema for this resource name
    :param dict channels: the channels
    :param list keys: the keys list for the instance of this resource
    :param str rsrc_name: override the resource name, defaults to meta data
    """
    _LOGGER.debug(f"keys: {keys}")

    # TODO: move to using "subscribe" in meta
    method = "get"
    rsrc_methods = schema.get("methods", [])

    if rsrc_methods and method not in rsrc_methods:
        return channels

    rsrc_descs = schema.get("descriptions", {})

    channels[baseurl] = get_channel(
        meta, baseurl, schema, rsrc_descs, keys=keys, rsrc_name=rsrc_name, is_list=True
    )

    return channels


def add_instance_level(
    meta: dict,
    baseurl: str,
    schema: dict,
    channels: dict,
    keys: list = None,
    rsrc_name: str = None,
):
    """Add instance level channels.

    :param dict meta: the meta data in the resource
    :param str baseurl: the baseurl to use for channels
    :param dict schema: the schema for this resource name
    :param dict channels: the channels
    :param list keys: the keys list for the instance of this resource
    :param str rsrc_name: override the resource name, defaults to meta data
    """
    _LOGGER.debug(f"keys: {keys}")

    # TODO: move to using "subscribe" in meta
    method = "get"
    rsrc_methods = schema.get("methods", [])

    if rsrc_methods and method not in rsrc_methods:
        return channels

    _LOGGER.debug(f"schema: {schema}")
    rsrc_inst_descs = schema["items"].get("descriptions", {})

    channels[baseurl] = get_channel(
        meta, baseurl, schema, rsrc_inst_descs, keys=keys, rsrc_name=rsrc_name
    )

    return channels


def add_instance_attrs_level(
    meta: dict,
    baseurl: str,
    schema: dict,
    channels: dict,
    keys: list = None,
    components: dict = None,
    rsrc_name: str = None,
):
    """Add the instance attr level channels.

    :param dict meta: the meta data in the resource
    :param str baseurl: the baseurl to use for channels
    :param dict schema: the schema for this resource name
    :param dict channels: the channels
    :param dict keys: the keys dict for the instance of this resource
    :param str rsrc_name: override the resource name, defaults to meta data
    """
    _LOGGER.debug(f"keys: {keys}")

    method = "get"
    rsrc_methods = schema.get("methods", [])
    rsrc_inst_descs = schema["items"].get("descriptions", {})

    for prop in schema["items"]["properties"]:
        path = "/".join([baseurl, prop])
        _LOGGER.debug(f"path: {path}")
        prop_schema = schema["items"]["properties"][prop]

        if "expose" in prop_schema and not prop_schema["expose"]:
            continue

        if rsrc_methods and method not in rsrc_methods:
            _LOGGER.info(
                f"Skipping the definition of {method} in resource instance attribute generation, "
                "as it is not in the defined methods requested"
            )
            continue

        # Recursively get channels for this property
        if "schema" in prop_schema:
            components["schemas"][prop] = prop_schema["schema"]["items"]
            _LOGGER.debug(f"components: {components}")
            get_channels(
                meta,
                path,
                prop_schema["schema"],
                keys=keys,
                rsrc_name=prop,
                channels=channels,
                components=components,
            )
        else:
            channels[path] = get_channel(
                meta,
                path,
                prop_schema,
                rsrc_inst_descs,
                keys=keys,
                rsrc_name=rsrc_name,
                attr_name=prop,
            )


def get_channels(
    meta: dict,
    baseurl: str,
    schema: dict,
    keys: list = None,
    rsrc_name: str = None,
    channels: dict = None,
    components: dict = None,
):
    """Get the channels, based on the resource definition."""
    if not channels:
        channels = {}
    _LOGGER.debug(f"keys: {keys}")

    asyncapi = meta.get("asyncapi", {})

    key = schema["key"]
    keys.append(key)

    # 1. Add resource level channels
    if asyncapi.get("channels", {}).get("resources"):
        add_resource_level(
            meta,
            baseurl,
            schema,
            channels,
            keys=keys,
            rsrc_name=rsrc_name,
        )

    if not asyncapi.get("channels", {}).get("instances") and not asyncapi.get("channels", {}).get(
        "instance_attrs"
    ):
        return channels

    if schema["type"] == "array" and not "key" in schema:
        raise spec_base.SchemaMissingAttribute(
            "You set asyncapi.channels.instances: true, but a 'key' is missing in schema {yaml.dump(schema)}"
        )

    instance_baseurl = "/".join([baseurl, f"{{{key['name']}}}"])
    _LOGGER.debug(f"instance_baseurl: {instance_baseurl}")

    # 2. Add instance level channels
    if asyncapi.get("channels", {}).get("instances"):

        add_instance_level(
            meta,
            instance_baseurl,
            schema,
            channels,
            keys=keys,
            rsrc_name=rsrc_name,
        )

    # 3. Add instance attribute level channels
    if asyncapi.get("channels", {}).get("instance_attrs"):
        add_instance_attrs_level(
            meta,
            instance_baseurl,
            schema,
            channels,
            keys=keys,
            components=components,
            rsrc_name=rsrc_name,
        )

    return channels


# pylint: disable=too-many-locals
def generate(rsrc_data: list, title: str, desc: str, summary: str, version: str):
    """Generate an AsyncAPI spec based on the resource data sent and other meta data."""
    components = {"schemas": {}}
    all_channels = {}
    servers = {}
    for rsrc in rsrc_data:
        rsrc_name = rsrc["name"]
        baseurl = "/"
        if "version_in_path" in rsrc and rsrc["version_in_path"]:
            baseurl += f"v{rsrc['version']}/"
        baseurl += rsrc_name
        _LOGGER.debug(f"baseurl: {baseurl}")

        # Extract and set high-level resource component schema
        rschema = rsrc["schema"]
        comp_schema = copy.deepcopy(rschema["items"])
        if "descriptions" in comp_schema:
            del comp_schema["descriptions"]
        components["schemas"][rsrc_name] = comp_schema

        meta = copy.deepcopy(rsrc)
        # TODO: remove
        if "schema" in meta:
            del meta["schema"]

        channels = get_channels(
            meta,
            baseurl,
            rschema,
            keys=[],
            channels={},
            components=components,
        )
        rsrc_servers = meta.get("asyncapi", {}).get("servers", {})
        servers.update(rsrc_servers)
        _LOGGER.debug(f"channels: {channels}")
        all_channels.update(channels)

    tmpl = spec_base.JINJA_ENV.get_template("asyncapi.jinja2")
    return tmpl.render(
        title=title,
        summary=summary,
        description=desc,
        version=version,
        servers=servers,
        components=components,
        channels=all_channels,
    )



"""
Generate OpenAPI 3.0 Spec
"""

# TODO: fix dupe code
# pylint: disable=duplicate-code

import copy
import http.client
import logging

from firestone.spec import _base as spec_base

DEFAULT_VERSION = "3.0.0"

# This is a list of all HTTP methods supported on high-level resource base
RSRC_HTTP_METHODS = ["delete", "get", "head", "patch", "post"]

# This is a list of all HTTP methods supported on an instance of a resource
RSRC_INST_HTTP_METHODS = ["delete", "get", "head", "patch", "put"]

# This is a list of all HTTP methods supported on attributes of an instance of a resource
RSRC_ATTR_HTTP_METHODS = ["delete", "get", "head", "put"]

_LOGGER = logging.getLogger(__name__)

# TODO add support for JSON Patch: https://www.jvt.me/posts/2022/05/29/openapi-json-patch/


def _dedup_params(params: list):
    seen = set()
    new_list = []
    for param in params:
        name = param["name"]
        if name not in seen:
            seen.add(name)
            new_list.append(param)

    return new_list


def get_responses(
    method: str,
    schema: dict,
    content_type: str,
    comp_name: str = None,
    attr_name: bool = None,
    is_list: bool = None,
):
    """Set schema for a given operation type."""
    if method == "head":
        return {"default": {"description": "Default HEAD response"}}

    resp_code_enum = http.client.OK
    if method == "post":
        resp_code_enum = http.client.CREATED

    responses = {
        resp_code_enum.value: {
            "description": f"Response for {resp_code_enum.name}",
            "content": {
                content_type: {
                    "schema": {},
                },
            },
        },
    }
    # Default to using the schema directly in the file
    schema_value = schema["items"] if "items" in schema and schema["type"] != "array" else schema

    # if a component name is provided, use that to reference it
    if comp_name:
        schema_value = f"#/components/schemas/{comp_name}"
    if attr_name:
        # schema_value = f"#/components/schemas/{comp_name}/properties/{attr_name}"
        schema_value = (
            schema["items"] if "items" in schema and schema["type"] != "array" else schema
        )
        responses[resp_code_enum.value]["content"][content_type]["schema"] = schema_value
        return responses

    responses[resp_code_enum.value]["content"][content_type]["schema"]["$ref"] = schema_value
    if is_list:
        responses[resp_code_enum.value]["content"][content_type]["schema"] = {
            "type": "array",
            "items": {"$ref": schema_value},
        }

    return responses


def _get_comp_name(rsrc_name: str, method: str):
    comp_name = rsrc_name if not rsrc_name.endswith("s") else rsrc_name[:-1]
    if method == "post":
        comp_name = f"Create{comp_name.capitalize()}"
    if method == "put":
        comp_name = f"Update{comp_name.capitalize()}"

    return comp_name


def get_method_op(
    path: str,
    method: str,
    schema: dict,
    desc: str = None,
    comp_name: str = None,
    attr_name: str = None,
    is_list: bool = None,
):
    """Get the specified method section for the paths."""
    if not desc:
        desc = f"{method} operation for {path}"
    content_type = spec_base.DEFAULT_CONTENT_TYPE
    opr = {
        "description": desc,
        "operationId": spec_base.get_opid(path, method),
    }

    # Now set the schema for responses
    _LOGGER.debug(f"method: {method}")
    _LOGGER.debug(f"schema: {schema}")
    _LOGGER.debug(f"comp_name: {comp_name}")
    _LOGGER.debug(f"attr_name: {attr_name}")
    _LOGGER.debug(f"is_list: {is_list}")
    opr["responses"] = get_responses(
        method,
        schema,
        content_type,
        comp_name=comp_name,
        attr_name=attr_name,
        is_list=is_list,
    )

    request_schema = None
    if method == "post":
        request_schema = (
            copy.deepcopy(schema["items"]) if "items" in schema else copy.deepcopy(schema)
        )
        if (
            http.client.CREATED in opr["responses"]
            and "$ref" in opr["responses"][http.client.CREATED]["content"][content_type]["schema"]
        ):
            request_schema = opr["responses"][http.client.CREATED]["content"][content_type][
                "schema"
            ]
    elif method == "put":
        _LOGGER.debug(f"Getting {method} operation")
        request_schema = copy.deepcopy(schema)
        _LOGGER.debug(f"request_schema: {request_schema}")
        if comp_name and not attr_name:
            request_schema = {"$ref": f"#/components/schemas/{comp_name}"}
    _LOGGER.debug(f"request_schema: {request_schema}")

    if request_schema:
        if "descriptions" in request_schema:
            del request_schema["descriptions"]
        if "key" in request_schema:
            del request_schema["key"]
        if "methods" in request_schema:
            del request_schema["methods"]
        if "query_params" in request_schema:
            del request_schema["query_params"]

        opr["requestBody"] = {
            "description": f"The request body for {path}",
            "required": True,
            "content": {
                spec_base.DEFAULT_CONTENT_TYPE: {
                    "schema": request_schema,
                },
            },
        }

    return opr


def get_params(
    baseurl: str,
    method: str,
    schema: dict,
    keys: list = None,
    param_schema: dict = None,
):
    """Get the parameters for this method."""
    parameters = []

    # Handle query params
    if "query_params" in schema:
        for param in copy.deepcopy(schema["query_params"]):
            _LOGGER.debug(f"param: {param}")
            methods = param.get("methods", [])
            _LOGGER.debug(f"methods: {methods}")
            if methods:
                del param["methods"]
            if methods and method not in methods:
                continue
            parameters.append(
                {
                    "name": param["name"],
                    "in": "query",
                    "required": param.get("required", False),
                    "schema": param.get("schema", {"type": "string"}),
                    "description": param.get("description"),
                }
            )

    # Handle path params
    _LOGGER.debug(f"keys: {keys}")
    if keys:
        for key in keys:
            key_name = key["name"]
            if key_name not in baseurl:
                continue
            parameters.append(
                {
                    "name": key_name,
                    "in": "path",
                    "required": True,
                    "schema": param_schema or {"type": "string"},
                }
            )

    return parameters


# pylint: disable=too-many-arguments
def add_resource_methods(
    rsrc_name: str,
    schema: dict,
    baseurl: str,
    paths: dict,
    methods: list = None,
    descs: dict = None,
    keys: list = None,
    default_query_params: dict = None,
    orig_rsrc_name: str = None,
):
    """Add resource level methods to the paths.

    :param str rsrc_name: the resource name
    :param dict schema: the schema for this resource name
    :param str baseurl: the baseurl to use for paths
    :param dict paths: the paths
    :param list keys: the keys for the instance of this resource
    :param dict default_query_params: the paths
    """
    if not descs:
        descs = {}

    paths[baseurl] = {}

    for method in RSRC_HTTP_METHODS:
        if methods and method not in methods:
            _LOGGER.info(
                f"Skipping the definition of {method} in resource generation, "
                "as it is not defined in methods.resource requested"
            )
            continue

        comp_name = _get_comp_name(rsrc_name, method)
        _LOGGER.debug(f"comp_name: {comp_name}")

        paths[baseurl][method] = get_method_op(
            baseurl,
            method,
            schema,
            desc=descs.get(method),
            comp_name=comp_name,
            is_list=(method == "get"),
        )
        paths[baseurl][method]["tags"] = [orig_rsrc_name or rsrc_name]

        # Add parameters
        params = get_params(baseurl, method, schema, keys=keys)
        if default_query_params:
            params.extend(default_query_params)
            params = _dedup_params(params)
        _LOGGER.debug(f"params: {params}")
        paths[baseurl][method]["parameters"] = params

    return paths


def add_instance_methods(
    rsrc_name: str,
    schema: dict,
    baseurl: str,
    paths: dict,
    methods: list = None,
    descs: dict = None,
    keys: list = None,
    orig_rsrc_name: str = None,
):
    """Add the instance methods to the paths.

    :param str rsrc_name: the resource name
    :param dict schema: the schema for this resource name
    :param str baseurl: the baseurl to use for paths
    :param list keys: the keys for the instance of this resource
    :param dict paths: the paths
    """
    if not descs:
        descs = {}

    paths[baseurl] = {}
    for method in RSRC_INST_HTTP_METHODS:
        _LOGGER.debug(f"baseurl: {baseurl}")
        if methods and method not in methods:
            _LOGGER.info(
                f"Skipping the definition of {method} in resource instance generation, "
                "as it is not defined in methods.instance requested"
            )
            continue

        comp_name = _get_comp_name(rsrc_name, method)
        _LOGGER.debug(f"comp_name: {comp_name}")

        paths[baseurl][method] = get_method_op(
            baseurl,
            method,
            schema,
            desc=descs.get(method),
            comp_name=comp_name,
        )

        # Add parameters
        params = get_params(baseurl, method, schema, keys=keys)
        _LOGGER.debug(f"params: {params}")
        paths[baseurl][method]["parameters"] = params

        # Add tags
        paths[baseurl][method]["tags"] = [orig_rsrc_name or rsrc_name]
        _LOGGER.debug(f"paths[baseurl][{method}]: {paths[baseurl][method]}")


# pylint: disable=too-many-locals
def add_instance_attr_methods(
    rsrc_name: str,
    schema: dict,
    baseurl: str,
    paths: dict,
    methods: dict = None,
    keys: list = None,
    default_query_params: dict = None,
    components: dict = None,
    orig_rsrc_name: str = None,
):
    """Add the instance attr methods to the paths.

    :param str rsrc_name: the resource name
    :param dict schema: the schema for this resource name
    :param str baseurl: the baseurl to use for paths
    :param list keys: the keys for the instance of this resource
    :param dict paths: the paths
    :param dict default_query_params: the paths
    """
    inst_methods = methods.get("instance_attrs", [])
    for prop in schema["items"]["properties"]:
        path = "/".join([baseurl, prop])
        _LOGGER.debug(f"path: {path}")
        prop_schema = schema["items"]["properties"][prop]

        if "expose" in prop_schema and not prop_schema["expose"]:
            continue

        paths[path] = {}
        for method in RSRC_ATTR_HTTP_METHODS:
            if inst_methods and method not in inst_methods:
                _LOGGER.info(
                    f"Skipping the definition of {method} in resource instance attribute generation, "
                    "as it is not in the defined methods requested"
                )
                continue

            comp_name = rsrc_name if not rsrc_name.endswith("s") else rsrc_name[:-1]
            inst_attr_op = get_method_op(
                path, method, prop_schema, comp_name=comp_name, attr_name=prop
            )
            _LOGGER.debug(f"inst_attr_op: {inst_attr_op}")

            paths[path][method] = inst_attr_op

            # Add parameters
            params = get_params(baseurl, method, schema, keys=keys)
            _LOGGER.debug(f"params: {params}")
            paths[path][method]["parameters"] = params

            # Add tags
            paths[path][method]["tags"] = [orig_rsrc_name or rsrc_name]
            _LOGGER.debug(f"paths[{path}][{method}]: {paths[path][method]}")

            # Recursively get paths for this property
            _LOGGER.debug(f"prop: {prop}")
            if "schema" in prop_schema:
                if "descriptions" in prop_schema["schema"]:
                    del prop_schema["schema"]["descriptions"]
                if "descriptions" in prop_schema["schema"]["items"]:
                    del prop_schema["schema"]["items"]["descriptions"]

                components = add_rsrc_components(components, prop, methods, prop_schema["schema"])
                components["schemas"][prop] = prop_schema["schema"]["items"]

                if (
                    rsrc_name in components["schemas"]
                    and prop in components["schemas"][rsrc_name]["properties"]
                ):
                    components["schemas"][rsrc_name]["properties"][prop] = {
                        "$ref": f"#/components/schemas/{prop}"
                    }

                get_paths(
                    prop,
                    prop_schema["schema"],
                    path,
                    paths,
                    keys=keys,
                    default_query_params=default_query_params,
                    components=components,
                    orig_rsrc_name=orig_rsrc_name,
                )


def get_paths(
    rsrc_name: str,
    rsrc: dict,
    baseurl: str,
    paths: dict = None,
    keys: list = None,
    default_query_params: dict = None,
    components: dict = None,
    orig_rsrc_name: str = None,
):
    """Get the paths for resource."""
    # Extract and set high-level resource component schema
    _LOGGER.debug(f"rsrc: {rsrc}")
    schema = rsrc["schema"] if "schema" in rsrc else rsrc
    methods = rsrc.get("methods", {})
    descs = rsrc.get("descriptions", {})

    if not paths:
        paths = {}

    if schema["type"] == "array" and schema["items"]["type"] == "object" and not "key" in schema:
        raise spec_base.SchemaMissingAttribute("A 'key' is missing in schema {yaml.dump(schema)}")

    key = None
    if "key" in schema:
        key = schema["key"]
        has_param = next((item for item in keys if item["name"] == key["name"]), None)
        _LOGGER.debug(f"has_param: {has_param}")
        if not has_param:
            keys.append(key)

    # 1. Add methods to high-level baseurl
    add_resource_methods(
        rsrc_name,
        schema,
        baseurl,
        paths,
        methods=methods.get("resource", {}),
        descs=descs.get("resource", {}),
        keys=keys,
        default_query_params=default_query_params,
        orig_rsrc_name=orig_rsrc_name,
    )
    _LOGGER.debug(f"paths[{baseurl}]: {paths[baseurl]}")

    instance_baseurl = "/".join([baseurl, f"{{{key['name']}}}"])
    _LOGGER.debug(f"instance_baseurl: {instance_baseurl}")

    # 2. Add paths for each attribute
    add_instance_methods(
        rsrc_name,
        schema,
        instance_baseurl,
        paths,
        methods=methods.get("instance", {}),
        descs=descs.get("instance", {}),
        keys=keys,
        orig_rsrc_name=orig_rsrc_name,
    )

    # 3. Add attribute path for instance of this resource
    add_instance_attr_methods(
        rsrc_name,
        schema,
        instance_baseurl,
        paths,
        methods=methods,
        keys=keys,
        default_query_params=default_query_params,
        components=components,
        orig_rsrc_name=orig_rsrc_name,
    )

    return paths


def add_rsrc_components(components: dict, rsrc_name: str, methods: dict, schema: dict):
    """Get the components for this resource."""
    comp_name = rsrc_name if not rsrc_name.endswith("s") else rsrc_name[:-1]

    # Reosurce level component, without required
    comp_schema = copy.deepcopy(schema["items"])
    if "descriptions" in comp_schema:
        del comp_schema["descriptions"]

    components["schemas"][comp_name] = comp_schema

    required = schema["items"].get("required", [])
    if required:
        del components["schemas"][comp_name]["required"]

    rscr_methods = methods.get("resource", [])
    rscr_inst_methods = methods.get("instance", [])
    _LOGGER.debug(f"rscr_methods: {rscr_methods}")
    _LOGGER.debug(f"rscr_inst_methods: {rscr_inst_methods}")

    # Create resource model
    if "post" in rscr_methods or "post" in rscr_inst_methods:
        create_key = f"Create{comp_name.capitalize()}"
        _LOGGER.info(f"Adding {create_key} to components")
        components["schemas"][create_key] = {
            "allOf": [
                {"$ref": f"#/components/schemas/{comp_name}"},
                {"type": "object"},
            ]
        }

        if required:
            components["schemas"][create_key]["allOf"][1]["required"] = required

    # Update resource model
    if "put" in rscr_methods or "put" in rscr_inst_methods:
        update_key = f"Update{comp_name.capitalize()}"
        _LOGGER.info(f"Adding {update_key} to components")
        components["schemas"][update_key] = {
            "allOf": [
                {"$ref": f"#/components/schemas/{comp_name}"},
                {"type": "object"},
            ]
        }

    return components


# pylint: disable=too-many-locals
def generate(
    rsrc_data: list,
    title: str,
    desc: str,
    summary: str,
    version: str,
    prefix: str = None,
    security: str = None,
    openapi_version: str = None,
):
    """Generate an OpenAPI spec based ont he resource data sent and other meta data."""
    components = {"schemas": {}}
    all_paths = {}
    for rsrc in rsrc_data:
        rsrc_name = rsrc["name"]
        baseurl = "/"
        if "version_in_path" in rsrc and rsrc["version_in_path"]:
            baseurl += f"v{rsrc['version']}/"
        baseurl += rsrc_name
        _LOGGER.debug(f"baseurl: {baseurl}")

        # Extract and set high-level resource component schema
        methods = rsrc.get("methods", {})
        components = add_rsrc_components(components, rsrc_name, methods, rsrc["schema"])
        _LOGGER.debug(f"components: {components['schemas']}")

        default_query_params = rsrc.get("default_query_params", [])
        _LOGGER.debug(f"default_query_params: {default_query_params}")

        paths = get_paths(
            rsrc["name"],
            rsrc,
            baseurl,
            paths={},
            keys=[],
            default_query_params=default_query_params,
            components=components,
            orig_rsrc_name=rsrc_name,
        )
        _LOGGER.debug(f"paths: {paths}")
        all_paths.update(paths)

    servers = []
    if prefix:
        servers.append({"url": prefix})

    security_name = None
    if security:
        security_name = security["name"]
        del security["name"]
        components["securitySchemes"] = {}
        components["securitySchemes"][security_name] = security

    tmpl = spec_base.JINJA_ENV.get_template("openapi.jinja2")
    return tmpl.render(
        title=title,
        summary=summary,
        description=desc,
        version=version,
        components=components,
        paths=all_paths,
        servers=servers,
        security_name=security_name,
        openapi_version=openapi_version,
    )



"""
Generate python Click based CLI from one or more resource schemas.
"""

import logging

from firestone.spec import _base as spec_base
from firestone.spec import openapi as spec_openapi

PARAM_TYPE_TO_ATTR_TYPE = {
    "string": "str",
    "integer": "int",
}

_LOGGER = logging.getLogger(__name__)


def _get_op_name(method: str, top_level: bool) -> str:
    if method == "get" and top_level:
        return "list"
    if method == "get":
        return "get"
    if method == "post":
        return "create"
    if method in ["put", "patch"]:
        return "update"
    if method == "delete":
        return method

    return None


def params_to_attrs(params: list, required: list = None, key_names: list = None):
    """Convert the params from OpenAPI spec to Click attributes."""
    if not required:
        required = []
    if not key_names:
        key_names = []
    _LOGGER.debug(f"params: {params}")
    _LOGGER.debug(f"key_names: {key_names}")

    attrs = []
    for param in params:
        _LOGGER.debug(f"param: {param}")
        param_schema = param.get("schema", param)
        param_type = param_schema.get("type", "string")
        cli_type = "str"

        # this means that this param/attribute is another object
        if param_type in ["object", "array"]:
            _LOGGER.info(f"{param['name']} is of type '{param_type}', processing special CLI type.")
            cli_type = "cli.FromJsonOrYaml()"
            items_type = param_schema.get("items", {}).get("type", "string")
            _LOGGER.debug(f"cli_type: {cli_type}")
            _LOGGER.debug(f"items_type: {items_type}")
            if param_type == "array" and items_type != "object":
                _LOGGER.info(
                    f"{param['name']} has items of type '{items_type}', setting click option to cli.StrList"
                )
                cli_type = "cli.StrList"
        elif "enum" in param_schema:
            _LOGGER.info(f"{param['name']} is of type '{param_type}', creating click.Choice()")
            enums = '","'.join(param_schema["enum"])
            cli_type = f'click.Choice(["{enums}"])'
        else:
            _LOGGER.info(
                f"{param['name']} is of type: {param_type}, looking up type in PARAM_TYPE_TO_ATTR_TYPE"
            )
            cli_type = PARAM_TYPE_TO_ATTR_TYPE.get(param_type)

        param_name = param["name"]
        required_val = param_name in required
        _LOGGER.debug(f"param_name: {param_name}")
        _LOGGER.debug(f"key_names: {key_names}")
        attrs.append(
            {
                "argument": param_name in key_names,
                "name": param_name,
                "description": param.get("description"),
                "type": cli_type,
                "required": param.get("required", required_val),
            }
        )
    return attrs


def get_resource_attrs(
    schema: dict, params: dict = None, check_required: bool = None, key_names: list = None
):
    """Get resource attributes."""
    _LOGGER.debug(f"key_names: {key_names}")
    props = schema.get("items", {}).get("properties", {})
    tmp_attrs = [{"name": attr, **(props[attr])} for attr in props]
    if params:
        tmp_attrs.extend(params)
    _LOGGER.debug(f"tmp_attrs: {tmp_attrs}")

    required = schema["items"].get("required", []) if check_required else []
    _LOGGER.debug(f"required: {required}")

    attrs = params_to_attrs(tmp_attrs, required, key_names=key_names)
    _LOGGER.debug(f"attrs: {attrs}")

    return attrs


def get_instance_ops(
    rsrc_name: str,
    schema: dict,
    baseurl: str,
    methods: list = None,
    descs: list = None,
    keys: list = None,
):
    """Add the instance methods to the paths.

    :param str rsrc_name: the resource name
    :param dict schema: the schema for this resource name
    :param str baseurl: the baseurl to use for paths
    :param list methods: optional set of methods to create for
    :param list keys: the keys for the instance of this resource
    :param dict paths: the paths
    """
    _LOGGER.debug(f"keys: {keys}")
    ops = []
    top_level = False
    for method in spec_openapi.RSRC_INST_HTTP_METHODS:
        if methods and method not in methods:
            _LOGGER.info(
                f"Skipping the definition of {method} in resource instance generation, "
                "as it is not in the defined methods requested"
            )
            continue

        op_name = _get_op_name(method, top_level)
        if not op_name:
            continue
        _LOGGER.info(f"Getting CLI attributes for {op_name}")

        op = {}
        op["name"] = op_name
        op["id"] = spec_base.get_opid(baseurl, method)
        op["description"] = descs.get(method, f"{op_name.capitalize()} operation for {rsrc_name}")

        params = spec_openapi.get_params(baseurl, method, schema, keys=keys)
        _LOGGER.debug(f"params: {params}")

        key_names = [key["name"] for key in keys]
        _LOGGER.debug(f"key_names: {key_names}")
        attrs = params_to_attrs(params, key_names=key_names)

        if op_name == "update":
            attrs = get_resource_attrs(schema, params=params, key_names=key_names)

        _LOGGER.debug(f"attrs: {attrs}")

        op["attrs"] = attrs

        ops.append(op)

    return ops


def get_resource_ops(
    rsrc_name: str,
    schema: dict,
    baseurl: str,
    methods: list = None,
    descs: list = None,
    keys: list = None,
    default_query_params: dict = None,
):
    """Add resource level methods to the ops.

    :param str rsrc_name: the resource name
    :param dict schema: the schema for this resource name
    :param str baseurl: the baseurl to use for ops
    :param list methods: optional set of methods to create for
    :param list keys: the keys for the instance of this resource
    :param dict default_query_params: the ops
    """
    _LOGGER.debug(f"keys: {keys}")
    ops = []

    top_level = True
    for method in spec_openapi.RSRC_HTTP_METHODS:
        if methods and method not in methods:
            _LOGGER.info(
                f"Skipping the definition of {method} in resource generation, "
                "as it is not in the defined methods requested"
            )
            continue

        op_name = _get_op_name(method, top_level)
        if not op_name:
            continue

        op = {}
        op["name"] = op_name
        op["id"] = spec_base.get_opid(baseurl, method)
        op["description"] = descs.get(method, f"{op_name.capitalize()} operation for {rsrc_name}")

        # Add params and attributes
        params = spec_openapi.get_params(baseurl, method, schema, keys=keys)
        if default_query_params:
            params.extend(default_query_params)
        _LOGGER.debug(f"params: {params}")

        attrs = params_to_attrs(params)

        if op_name == "create":
            check_required = True
            attrs = get_resource_attrs(schema, check_required=check_required)
        _LOGGER.debug(f"attrs: {attrs}")

        op["attrs"] = attrs

        ops.append(op)

    return ops


def get_ops(
    rsrc: dict,
    baseurl: str,
    ops: dict = None,
    keys: list = None,
    default_query_params: dict = None,
):
    """Get the operations for this resource."""
    rsrc_name = rsrc["name"]
    schema = rsrc["schema"]
    methods = rsrc.get("methods", {})
    descs = rsrc.get("descriptions", {})

    if not ops:
        ops = {}

    if schema["type"] == "array" and not "key" in schema:
        raise spec_base.SchemaMissingAttribute("A 'key' is missing in schema {yaml.dump(schema)}")

    key = None
    if "key" in schema:
        key = schema["key"]
        has_param = next((item for item in keys if item["name"] == key["name"]), None)
        _LOGGER.debug(f"has_param: {has_param}")
        if not has_param:
            keys.append(key)

    _LOGGER.debug(f"keys: {keys}")

    # 1. Add operations to high-level baseurl
    ops["resource"] = get_resource_ops(
        rsrc_name,
        schema,
        baseurl,
        methods=methods.get("resource", {}),
        descs=descs.get("resource", {}),
        keys=keys,
        default_query_params=default_query_params,
    )

    instance_baseurl = "/".join([baseurl, f"{{{key['name']}}}"])
    _LOGGER.debug(f"instance_baseurl: {instance_baseurl}")

    # 2. Get instance operations
    ops["instance"] = get_instance_ops(
        rsrc_name,
        schema,
        instance_baseurl,
        methods=methods.get("instance", {}),
        descs=descs.get("instance", {}),
        keys=keys,
    )

    return ops


# pylint: disable=too-many-locals
def generate(
    pkg: str, client_pkg: str, rsrc_data: list, title: str, desc: str, summary: str, version: str
):
    """Generate a Click based CLI script based on the resource data sent and other meta data."""
    rsrcs = []
    for rsrc in rsrc_data:
        rsrc_name = rsrc["name"]
        baseurl = "/"
        if "version_in_path" in rsrc and rsrc["version_in_path"]:
            baseurl += f"v{rsrc['version']}/"
        baseurl += rsrc_name
        _LOGGER.debug(f"baseurl: {baseurl}")

        default_query_params = rsrc.get("default_query_params", [])
        _LOGGER.debug(f"default_query_params: {default_query_params}")

        ops = get_ops(
            rsrc,
            baseurl,
            ops={},
            keys=[],
            default_query_params=default_query_params,
        )
        _LOGGER.debug(f"ops: {ops}")
        rsrcs.append(
            {
                "name": rsrc_name,
                "operations": ops,
            }
        )

    _LOGGER.info(f"rsrcs: {rsrcs}")
    tmpl = spec_base.JINJA_ENV.get_template("main.py.jinja2")
    return tmpl.render(
        title=title,
        summary=summary,
        description=desc,
        version=version,
        pkg=pkg,
        client_pkg=client_pkg,
        rsrcs=rsrcs,
    )
